@@ -3,7 +3,7 @@
 # .. we'll use this to pass it to the child script ..
 _clean_globals = globals().copy()
 
-__version__ = '0.28'
+__version__ = '0.30'
 
 _CMD_USAGE = "python -m memory_profiler script_file.py"
 
@@ -66,7 +66,8 @@ def _get_memory(pid, timestamps=False, include_children=False):
         # .. subprocess.check_output appeared in 2.7, using Popen ..
         # .. for backwards compatibility ..
         out = subprocess.Popen(['ps', 'v', '-p', str(pid)],
-              stdout=subprocess.PIPE).communicate()[0].split(b'\n')
+                               stdout=subprocess.PIPE
+                               ).communicate()[0].split(b'\n')
         try:
             vsz_index = out[0].split().index(b'RSS')
             mem = float(out[1].split()[vsz_index]) / 1024
@@ -84,7 +85,7 @@ def _get_memory(pid, timestamps=False, include_children=False):
                                   'platforms')
 
 
-class Timer(Process):
+class MemTimer(Process):
     """
     Fetch memory consumption from over a time interval
     """
@@ -95,6 +96,7 @@ class Timer(Process):
         self.pipe = pipe
         self.cont = True
         self.max_usage = max_usage
+        self.n_measurements = 1
 
         if "timestamps" in kw:
             self.timestamps = kw["timestamps"]
@@ -106,25 +108,31 @@ class Timer(Process):
             del kw["include_children"]
         else:
             self.include_children = False
+        # get baseline memory usage
+        self.mem_usage = [
+            _get_memory(self.monitor_pid, timestamps=self.timestamps,
+                        include_children=self.include_children)]
+        super(MemTimer, self).__init__(*args, **kw)
 
-        super(Timer, self).__init__(*args, **kw)
 
     def run(self):
-        m = _get_memory(self.monitor_pid, timestamps=self.timestamps,
-                        include_children=self.include_children)
-        if not self.max_usage:
-            timings = [m]
-        else:
-            timings = m
         self.pipe.send(0)  # we're ready
-        while not self.pipe.poll(self.interval):
-            m = _get_memory(self.monitor_pid, timestamps=self.timestamps,
+        stop = False
+        while True:
+            cur_mem = _get_memory(self.monitor_pid, timestamps=self.timestamps,
                             include_children=self.include_children)
             if not self.max_usage:
-                timings.append(m)
+                self.mem_usage.append(cur_mem)
             else:
-                timings = max([m, timings])
-        self.pipe.send(timings)
+                self.mem_usage[0] = max(cur_mem, self.mem_usage[0])
+            self.n_measurements += 1
+            if stop:
+                break
+            stop = self.pipe.poll(self.interval)
+            # do one more iteration
+
+        self.pipe.send(self.mem_usage)
+        self.pipe.send(self.n_measurements)
 
 
 def memory_usage(proc=-1, interval=.1, timeout=None, timestamps=False,
@@ -170,6 +178,8 @@ def memory_usage(proc=-1, interval=.1, timeout=None, timestamps=False,
     -------
     mem_usage : list of floating-poing values
         memory usage, in MiB. It's length is always < timeout / interval
+        if max_usage is given, returns the two elements maximum memory and
+        number of measurements effectuated
     ret : return value of the profiled function
         Only returned if retval is set to True
     """
@@ -211,17 +221,22 @@ def memory_usage(proc=-1, interval=.1, timeout=None, timestamps=False,
             raise ValueError('Function expects %s value(s) but %s where given'
                              % (n_args, len(args)))
 
-        child_conn, parent_conn = Pipe()  # this will store Timer's results
-        p = Timer(os.getpid(), interval, child_conn, timestamps=timestamps,
-                  max_usage=max_usage)
-        p.start()
-        parent_conn.recv()  # wait until we start getting memory
-        returned = f(*args, **kw)
-        parent_conn.send(0)  # finish timing
-        ret = parent_conn.recv()
-        if retval:
-            ret = ret, returned
-        p.join(5 * interval)
+        while True:
+            child_conn, parent_conn = Pipe()  # this will store MemTimer's results
+            p = MemTimer(os.getpid(), interval, child_conn, timestamps=timestamps,
+                      max_usage=max_usage)
+            p.start()
+            parent_conn.recv()  # wait until we start getting memory
+            returned = f(*args, **kw)
+            parent_conn.send(0)  # finish timing
+            ret = parent_conn.recv()
+            n_measurements = parent_conn.recv()
+            if retval:
+                ret = ret, returned
+            p.join(5 * interval)
+            if n_measurements > 4 or interval < 1e-6:
+                break
+            interval /= 10.
     elif isinstance(proc, subprocess.Popen):
         # external process, launched from Python
         line_count = 0
@@ -375,14 +390,14 @@ class TimeStamper:
                     (function_name,) + ts[0] + ts[1]))
 
 
-class LineProfiler:
+class LineProfiler(object):
     """ A profiler that records the amount of memory for each line """
 
     def __init__(self, **kw):
-        self.functions = list()
         self.code_map = {}
         self.enable_count = 0
         self.max_mem = kw.get('max_mem', None)
+        self.prevline = None
 
     def __call__(self, func):
         self.add_function(func)
@@ -406,7 +421,6 @@ class LineProfiler:
             return
         if code not in self.code_map:
             self.code_map[code] = {}
-            self.functions.append(func)
 
     def wrap_function(self, func):
         """ Wrap a function to profile it.
@@ -424,6 +438,7 @@ class LineProfiler:
     def run(self, cmd):
         """ Profile a single executable statement in the main namespace.
         """
+        # TODO: can this be removed ?
         import __main__
         main_dict = __main__.__dict__
         return self.runctx(cmd, main_dict, main_dict)
@@ -456,13 +471,15 @@ class LineProfiler:
 
     def trace_memory_usage(self, frame, event, arg):
         """Callback for sys.settrace"""
-        if event in ('line', 'return') and frame.f_code in self.code_map:
-            lineno = frame.f_lineno
-            if event == 'return':
-                lineno += 1
-            entry = self.code_map[frame.f_code].setdefault(lineno, [])
-            entry.append(_get_memory(-1))
-
+        if (event in ('call', 'line', 'return')
+                and frame.f_code in self.code_map):
+            if event != 'call':
+                # "call" event just saves the lineno but not the memory
+                mem = _get_memory(-1)
+                # if there is already a measurement for that line get the max
+                old_mem = self.code_map[frame.f_code].get(self.prevline, 0)
+                self.code_map[frame.f_code][self.prevline] = max(mem, old_mem)
+            self.prevline = frame.f_lineno
         return self.trace_memory_usage
 
     def trace_max_mem(self, frame, event, arg):
@@ -525,44 +542,27 @@ def show_results(prof, stream=None, precision=1):
             continue
         all_lines = linecache.getlines(filename)
         sub_lines = inspect.getblock(all_lines[code.co_firstlineno - 1:])
-        linenos = range(code.co_firstlineno, code.co_firstlineno +
-                        len(sub_lines))
-        lines_normalized = {}
+        linenos = range(code.co_firstlineno,
+                        code.co_firstlineno + len(sub_lines))
 
         header = template.format('Line #', 'Mem usage', 'Increment',
                                  'Line Contents')
         stream.write(header + '\n')
         stream.write('=' * len(header) + '\n')
-        # move everything one frame up
-        keys = sorted(lines.keys())
 
-        k_old = keys[0] - 1
-        lines_normalized[keys[0] - 1] = lines[keys[0]]
-        for i in range(1, len(lines_normalized[keys[0] - 1])):
-            lines_normalized[keys[0] - 1][i] = -1.
-        k = keys.pop(0)
-        while keys:
-            lines_normalized[k] = lines[keys[0]]
-            for i in range(len(lines_normalized[k_old]),
-                           len(lines_normalized[k])):
-                lines_normalized[k][i] = -1.
-            k_old = k
-            k = keys.pop(0)
-
-        first_line = sorted(lines_normalized.keys())[0]
-        mem_old = max(lines_normalized[first_line])
-        precision = int(precision)
-        template_mem = '{{0:{0}.{1}'.format(precision + 4, precision) + 'f} MiB'
-        for i, l in enumerate(linenos):
+        mem_old = lines[min(lines.keys())]
+        float_format = '{0}.{1}f'.format(precision + 4, precision)
+        template_mem = '{0:' + float_format + '} MiB'
+        for line in linenos:
             mem = ''
             inc = ''
-            if l in lines_normalized:
-                mem = max(lines_normalized[l])
+            if line in lines:
+                mem = lines[line]
                 inc = mem - mem_old
                 mem_old = mem
                 mem = template_mem.format(mem)
                 inc = template_mem.format(inc)
-            stream.write(template.format(l, mem, inc, sub_lines[i]))
+            stream.write(template.format(line, mem, inc, all_lines[line - 1]))
         stream.write('\n\n')
 
 
@@ -629,7 +629,7 @@ def magic_mprun(self, parameter_s=''):
             funcs.append(eval(name, global_ns, local_ns))
         except Exception as e:
             raise UsageError('Could not find function %r.\n%s: %s' % (name,
-                e.__class__.__name__, e))
+                             e.__class__.__name__, e))
 
     profile = LineProfiler()
     for func in funcs:
@@ -657,7 +657,7 @@ def magic_mprun(self, parameter_s=''):
             message = "*** SystemExit exception caught in code being profiled."
         except KeyboardInterrupt:
             message = ("*** KeyboardInterrupt exception caught in code being "
-                "profiled.")
+                       "profiled.")
     finally:
         if had_profile:
             builtins.__dict__['profile'] = old_profile
@@ -736,16 +736,23 @@ def magic_memit(self, line=''):
         timeout = None
     interval = float(getattr(opts, 'i', 0.1))
 
-    mem_usage = []
+    # I've noticed we get less noisier measurements if we run
+    # a garbage collection first
+    import gc
+    gc.collect()
+
+    mem_usage = 0
     counter = 0
+    baseline = memory_usage()[0]
     while counter < repeat:
         counter += 1
         tmp = memory_usage((_func_exec, (stmt, self.shell.user_ns)),
-                           timeout=timeout, interval=interval)
-        mem_usage.extend(tmp)
+                           timeout=timeout, interval=interval, max_usage=True)
+        mem_usage = max(mem_usage, tmp[0])
 
     if mem_usage:
-        print('maximum of %d: %f MiB per loop' % (repeat, max(mem_usage)))
+        print('peak memory: %.02f MiB, increment: %.02f MiB' %
+              (mem_usage, mem_usage - baseline))
     else:
         print('ERROR: could not read memory usage, try with a lower interval '
               'or more iterations')
@@ -773,10 +780,12 @@ if __name__ == '__main__':
     from optparse import OptionParser
     parser = OptionParser(usage=_CMD_USAGE, version=__version__)
     parser.disable_interspersed_args()
-    parser.add_option("--pdb-mmem", dest="max_mem", metavar="MAXMEM",
+    parser.add_option(
+        "--pdb-mmem", dest="max_mem", metavar="MAXMEM",
         type="float", action="store",
         help="step into the debugger when memory exceeds MAXMEM")
-    parser.add_option('--precision', dest="precision", type="int",
+    parser.add_option(
+        '--precision', dest="precision", type="int",
         action="store", default=3,
         help="precision of memory output in number of significant digits")
     parser.add_option("-o", dest="out_filename", type="str",
@@ -802,7 +811,7 @@ if __name__ == '__main__':
     try:
         if sys.version_info[0] < 3:
             # we need to ovewrite the builtins to have profile
-            # globally defined (global variables is not enought
+            # globally defined (global variables is not enough
             # for all cases, e.g. a script that imports another
             # script where @profile is used)
             import __builtin__
